//***********************************************************************
//Application design
//------------------
//To support GUI access, we use two threads: The GUI thread and the Processing thread.
//The main thread is the GUI thread. If the GUI thread invokes a method, such as
//connect, a message is sent to the processing thread using the postmessage windows
//command.
//If an event is invoked by the processingThread e.g.: onMessageReceived, it is
//synchronized into the main thread using the Thread.Synchronize method.
//***********************************************************************

unit unSMSClient;

interface
uses windows,messages,Classes,SysUtils,unSMSMessage,unSocketReader, unOzTimer,
     unProcessingThread,unErrorCodes,IdBaseComponent, IdComponent,
     IdTCPConnection,idStackConsts,IdTCPClient,IdGlobal, unWideHash, StrUtils, IdURI;

Const
   //****************************************
   // Messages used for synchonization
   //****************************************
   WM_OZ_STOP = WM_USER + 200;
   WM_TCP_SOCKET_CONNECTED = WM_USER+201;
   WM_TCP_SOCKET_DISCONNECTED = WM_USER+202;
   WM_TCP_DATAAVAILABLE = WM_USER+203;
   WM_CLIENT_CONNECTING = WM_USER+208;
   WM_CLIENT_LINKUP = WM_USER+209;
   WM_CLIENT_CONNECTED = WM_USER+210;
   WM_CLIENT_MESSAGEDELIVERY_ERROR = WM_USER+211;
   WM_CLIENT_KEEPALIVE = WM_USER+212;
   WM_DO_LOGIN = WM_USER+213;
   WM_COMMUNICATION = WM_USER+214;
   WM_CONNECT = WM_USER+215;
   BufferSize = 50000;

   //****************************************
   // TCP protocol
   //****************************************
   stx : char = chr(2);
   etx : char = chr(3);



Type

   TDirection = (Inbound,Outbound);

   TSimpleEvent = procedure(Sender : TObject) of object;
   TErrorEvent = procedure(Sender : TObject; ErrorCode: integer; ErrorMessage:string) of object;
   TDeliveryEvent = procedure(
      Sender : TObject;
      Messageid : string;
      Senderaddress : widestring;
      Receiver : widestring;
      Messagedata : widestring;
      Messagetype : string;
      Serviceprovider : string;
      Sentdate : string;
      Receiveddate : string
      ) of object;
   TDeliveryErrorEvent = procedure(
      Sender : TObject;
      Messageid : string;
      Senderaddress : widestring;
      Receiver : widestring;
      Messagedata : widestring;
      Messagetype : string;
      Serviceprovider : string;
      Sentdate : string;
      Receiveddate : string;
      ErrorCode:integer;
      ErrorMessage:widestring) of object;
   TCommunicationEvent = procedure(
      Sender : TObject;
      direction : TDirection;
      data : string) of object;
   TConnectingEvent = procedure(
      Sender : TObject;
      host : string;
      port : integer) of object;

   TSMSClient = class(TComponent)
      private
        transactionID : integer;
        fHost : string;
        fPort : integer;
        fUsername : string;
        fPassword : string;
        fLastErrorCode : integer;
        fLastErrorMessage : string;
        fConnected : boolean;
        fSocketTimeout : integer; //in seconds
        fKeepalivePeriod : integer; //in seconds

        tcpClient : TIDTcpClient;
        tcpInputBuffer : string;
        socketThread : TSocketReaderThread;
        processingThread : TProcessingThread;
        timerKeepAlive : TOzTimer;

        //*** Buffer
        smsBuffer  : Array[0..BufferSize+2] of TSMSMessage;
        bufferStart:Int32;
        bufferStop:Int32;
        //*** Buffer

        //Published events
        fonClientConnected : TSimpleEvent;
        fonClientDisconnected : TSimpleEvent;
        fonClientConnectionError : TErrorEvent;
        fonMessageAcceptedForDelivery : TDeliveryEvent;
        fonMessageDeliveredToNetwork : TDeliveryEvent;
        fonMessageDeliveredToHandset : TDeliveryEvent;
        fonMessageDeliveryError : TDeliveryErrorEvent;
        fonMessageReceived : TDeliveryEvent;
        fonCommunication : TCommunicationEvent;
        fonConnecting : TConnectingEvent;
        fonSocketConnected : TConnectingEvent;

        //Client events
        procedure onIncomingPDU(inputPDU : string);


        //Socket management
        procedure onTCPClientConnected(Sender: TObject);
        procedure onTCPClientDisConnected(Sender: TObject);
        procedure onTCPClientStatus(axSender: TObject;
                 const axStatus: TIdStatus; const asStatusText: String);

        //Thread management - to receive incoming messages
        procedure startThread;
        procedure stopThread;

        //Event handler wrappers
        procedure doOnClientConnected;
        procedure doOnClientDisconnected;
        procedure doonClientConnectionError(errorCode: integer; errorMessage:string);
        procedure doOnMessageAcceptedForDelivery(sms : TSMSMessage);
        procedure doOnMessageDeliveredToNetwork(sms : TSMSMessage);
        procedure doOnMessageDeliveredToHandset(sms : TSMSMessage);
        procedure doOnMessageDeliveryError(sms : TSMSMessage);
        procedure doOnMessageReceived(sms : TSMSMessage);

        //Support functions
        procedure versioncheck(var continueLogin : boolean);
        procedure login(var continuelogin : boolean);
        function sendPDU(pdu : string) : boolean;
        function waitforPDU(var pduReceived : string; timeout:integer):boolean;
        function readPDU(var pduReceived : string; var count:integer) : boolean;
        function ozUrlEncode(data : widestring) : string;
        function DateTimeToUnix(ConvDate: TDateTime): Longint;
        function unixnow : int64;
        function visualize(s : string):string;
        function randomStringL(len : integer):string;
        function getNextTransID : string;
        procedure onKeepalive(Sender: TObject);
        procedure sendKeepalive;
        procedure doConnect(host: string; port : integer; async: boolean);
        Procedure doLogin;
        procedure connect(value : boolean);
        procedure disconnect;
        procedure logCom(direction : TDirection; data : string);
        function getMsgHandle : HWND;
        procedure fSynchronized(Value : boolean);
      public
        property msgHandle : HWND read getMsgHandle;

        //Published methods
        constructor create(AOwner: TComponent); override;
        destructor destroy; override;
        procedure beginAsyncConnect;
        function sendMessage(receiver : widestring; messagedata : widestring) : string;
        function sendMessageEx(sender : widestring; receiver : widestring;
            messagedata : widestring; messagetype : string; serviceprovider : string) : string;
        function getLastErrorMessage : string;
        function getLastErrorCode : integer;

        //Support methods
        procedure wndProc(var Msg: TMessage);

        published
        //Published properties
        property Connected : boolean read fConnected write connect;
        property Host : string read fHost write fHost;
        property Port : integer read fPort write fPort;
        property Username : string read fUserName write fUserName;
        property Password : string read fPassword write fPassword;
        property KeepalivePeriod : integer read fKeepalivePeriod write fKeepalivePeriod;
        property SocketTimeout : integer read fSocketTimeout write fSocketTimeout;
        property Synchronized : boolean write fSynchronized;

        //Published events
        property onClientConnected : TSimpleEvent read fonClientConnected write fonClientConnected;
        property onClientDisconnected : TSimpleEvent read fonClientDisconnected write fonClientDisconnected;
        property onClientConnectionError : TErrorEvent read fonClientConnectionError write fonClientConnectionError;
        property onMessageAcceptedForDelivery : TDeliveryEvent read fonMessageAcceptedForDelivery write fonMessageAcceptedForDelivery;
        property onMessageDeliveredToNetwork : TDeliveryEvent read fonMessageDeliveredToNetwork write fonMessageDeliveredToNetwork;
        property onMessageDeliveredToHandset : TDeliveryEvent read fonMessageDeliveredToHandset write fonMessageDeliveredToHandset;
        property onMessageDeliveryError : TDeliveryErrorEvent read fonMessageDeliveryError write fonMessageDeliveryError;
        property onMessageReceived : TDeliveryEvent read fonMessageReceived write fonMessageReceived;
        property onLowLevelCommunication : TCommunicationEvent read fonCommunication write fonCommunication;
        property onConnecting : TConnectingEvent read fonConnecting write fonConnecting;
        property onSocketConnected : TConnectingEvent read fonSocketConnected write fonSocketConnected;
   end;

   procedure Register;



implementation

{$R unSMSClient.dcr}


constructor TSMSClient.create(AOwner: TComponent);
Begin
   inherited create(AOwner);
   Randomize;
   fConnected     := false;
   fPort          := 9500;
   fHost          := '127.0.0.1';
   fUsername      := 'admin';
   fPassword      := 'abc123';
   fSocketTimeout := 30;
   transactionID  := -1;
   socketThread   := nil;
   fKeepAlivePeriod := 60;
   timerKeepalive := TozTimer.Create;
   timerKeepalive.enabled := false;
   timerKeepalive.onTimer := onKeepalive;
   startThread;
End;

destructor TSMSClient.destroy;
Begin
   if fConnected then disconnect;
   stopThread;
   timerKeepAlive.free;
   inherited destroy;
End;

//***********************************************************************
//SMS Client methods
//***********************************************************************
procedure TSMSClient.connect(Value : boolean);
Begin
   if (Value) then
    doConnect(fHost,fPort, false)
   else
    disconnect;
End;

procedure TSMSClient.beginAsyncConnect;
Begin
   doConnect(fHost,fPort, true);
End;

procedure TSMSClient.doConnect(host: string; port : integer; async : boolean);
var
t:Tthread;
Begin
   socketThread := nil;
   transactionID := -1;

   if (tcpClient <> nil) then begin
       if (tcpClient.Connected) then tcpClient.Disconnect;
   end else begin
       //Create socket
       tcpClient := TIDTcpClient.create(nil);
   end;

   fHost := host;
   fPort := port;

   if (assigned(fOnConnecting)) then begin
   try
      fOnConnecting(self,fhost,fport);
   except
   end;
   end;

   tcpClient.onConnected    := onTCPClientConnected;
   tcpClient.onDisConnected := onTCPClientDisConnected;
   tcpClient.onStatus       := onTCPClientStatus;
   tcpClient.Host           := host;
   tcpClient.Port           := port;
   tcpInputBuffer           := '';

   try
      tcpClient.Connect;
      fConnected := true;

      //*** BufferReadStart

      bufferStop:=0;
      bufferStart:=0;
        t.CreateAnonymousThread(
        procedure begin
         While true do
           begin
           if bufferStart = bufferStop then
             begin
             if bufferStop>0 then
                 bufferStop:= bufferStop;

                 sleep(200);
              end
            else
            begin
              if(bufferStart=BufferSize-2) then
                  begin
                    bufferStart:=0;
                  end
                else
                  begin
                    bufferStart:=bufferStart+1;
                  end;

                TThread.Synchronize(nil, procedure
                  begin
                    doOnMessageReceived(smsBuffer[bufferStart]);
                  end);
                  sleep(50);
            end
            end
        end
        ).Start;


      //***


      postMessage(processingThread.msgHandle,WM_CLIENT_LINKUP,0,0);
   except on e:exception do
      begin
        fLastErrorCode := ERROR_SOCKET_CONNECT;
        fLastErrorMessage := e.Message;
        doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
      end;
   end;

   if (fConnected) then begin
      if (async) then begin
          postMessage(processingThread.msgHandle,WM_DO_LOGIN,0,0);
      end else begin
          doLogin;
      end;
   end;
end;

Procedure TSMSClient.doLogin;
var
   continueLogin : boolean;
Begin
   if fConnected then versioncheck(continueLogin);
   if (continueLogin) then begin
       if fConnected then login(continueLogin);
       if (continueLogin) then begin
           if fConnected then begin
               postMessage(processingThread.msghandle,WM_CLIENT_CONNECTED,0,0);
           end;
           if fConnected then begin
               postMessage(processingThread.msgHandle,WM_TCP_DATAAVAILABLE,0,0);
               socketThread := TSocketReaderThread.Create(true);
               socketThread.msgHandle := processingThread.msgHandle;
               socketThread.socket := tcpClient;
               socketThread.Resume;
           end;

           if fConnected then begin
             //Create keepalive timer
             timerKeepalive.enabled        := false;
             timerKeepalive.interval       := fKeepAlivePeriod*1000;
             timerKeepalive.tag            := 0;
             timerKeepalive.enabled        := true;
           end;
       end;
   end;
End;

procedure TSMSClient.versioncheck(var continueLogin : boolean);
var
   versioncheckpdu : string;
   versioncheckresppdu : string;
   pduhash : TozWideHash;
   action : string;
   errorcodestr : string;
   redirectHost : string;
   redirectPort : string;
Begin
   continueLogin := true;
   //Version check PDU
   versioncheckpdu := stx+
      'transactionid='+getNextTransID+'&'+
      'action=getversion'+
      '&responseformat=urlencoded'+
      etx;

   //Send version check PDU and process response PDU
   if (not sendPDU(versioncheckpdu)) then begin
       doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
   end else if (not waitforPDU(versioncheckresppdu,fSocketTimeout)) then begin
       doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
   end else begin
       pduhash := tozWideHash.Create;
       pduhash.fromUrl(versioncheckresppdu);
       action := pduhash.get('action');
       if (action='redirect') then begin
          redirectHost := pduhash.get('redirecthost');
          redirectPort := pduhash.get('redirectport');
          doConnect(redirectHost,strtoint(redirectPort),false);
          continueLogin := false;
       end else if (action='error') then begin
          errorcodestr := pduhash.get('errorcode');
          fLastErrorCode := strtoint(errorcodestr);
          fLastErrorMessage := pduhash.get('errormessage');
          doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
       end else if (action='getversion') then begin
          //all protocol versions are supported
       end else begin
          fLastErrorCode := ERROR_PDU_UNEXPECTED;
          fLastErrorMessage := 'Unexpected response received from server. '+visualize(versioncheckresppdu);
          doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
       end;
       pduhash.free;
   end;
End;

procedure TSMSClient.login(var continuelogin : boolean);
var
   loginpdu : string;
   loginresppdu : string;
   pduhash : TozWideHash;
   action : string;
   errorcodestr : string;
   redirecthost : string;
   redirectport : string;
Begin
   continuelogin := true;
   //Build login PDU
   loginpdu := stx+
      'transactionid='+getNextTransID+'&'+
      'action=login'+
      '&username='+ozUrlencode(fusername)+
      '&password='+ozUrlencode(fpassword)+
      '&responseformat=urlencoded'+etx;

   //Send login PDU and process response PDU
   if (not sendPDU(loginpdu)) then begin
       doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
   end else if (not waitforPDU(loginresppdu,fSocketTimeout)) then begin
       doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
   end else begin
       pduhash := tozWideHash.Create;
       pduhash.fromUrl(loginresppdu);
       action := pduhash.get('action');
       if (action='redirect') then begin
          redirectHost := pduhash.get('redirecthost');
          redirectPort := pduhash.get('redirectport');
          doConnect(redirectHost,strtoint(redirectPort),false);
          continueLogin := false;
       end else if (action='error') then begin
          errorcodestr := pduhash.get('errorcode');
          fLastErrorCode := strtoint(errorcodestr);
          fLastErrorMessage := pduhash.get('errormessage');
          doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
       end;
       pduhash.free;       
   end;

End;

procedure TSMSClient.disconnect;
Begin
   timerKeepAlive.enabled := false;
   try
      if (tcpClient<>nil) then begin
      if tcpClient.Connected then tcpClient.Disconnect;
      if assigned(socketThread) then begin
         socketThread.WaitFor;
         socketThread.free;
         socketThread := nil;
      end;
      end;
   except
   end;
   fConnected := false;
End;

procedure TSMSClient.fSynchronized(Value : boolean);
Begin
    processingThread.synchronized := Value;
End;


function TSMSClient.sendMessage(receiver : widestring; messagedata : widestring) : string;
Begin
    Result := sendMessageEx('',receiver,messagedata,'SMS:TEXT','');
End;

function TSMSClient.sendMessageEx(
      sender : widestring;
      receiver : widestring;
      messagedata : widestring;
      messagetype : string;
      serviceprovider : string
      ) : string;
var
    messageID : string;
    submitPDU : string;
    sms : TSMSMessage;
Begin

    messageID := randomStringL(8);

    submitPDU := stx+
      'transactionid='+getNextTransID+'&'+
      'action=sendmessage'+
      '&responseformat=urlencoded'+
      '&originator='+ozUrlencode(sender)+
      '&recipient='+ozUrlencode(receiver)+
      '&messagetype='+ozUrlencode(messagetype)+
      '&messagedata='+ozUrlencode(messagedata)+
      '&messageid='+ozUrlencode(messageID)+
      '&serviceprovider='+ozUrlencode(serviceprovider)+
      etx;

    if (not sendPDU(submitPDU)) then begin
        sms := TSMSMessage.Create;
        sms.messageID := messageID;
        sms.sender := sender;
        sms.receiver := receiver;
        sms.messagedata := messagedata;
        sms.messagetype := messagetype;
        sms.serviceprovider := serviceprovider;
        sms.errorcode := ERROR_SOCKET_SEND;
        sms.errormessage := 'Client disconnected from SMS gateway during message submission.';
        postMessage(processingThread.msghandle,WM_CLIENT_MESSAGEDELIVERY_ERROR,integer(sms),0);
    end;

    Result := messageID;
End;

procedure TSMSClient.sendKeepalive;
var
    keepalivePDU : string;
Begin
    keepalivePDU := stx+
    'transactionid='+getNextTransID+'&'+
    'action=keepalive'+
    '&responseformat=urlencoded'+
    etx;

    if (not sendPDU(keepalivePDU)) then begin
         fLastErrorCode := ERROR_KEEPALIVE;
         fLastErrorMessage := 'Keepalive failed.';
         doOnClientConnectionError(fLastErrorCode,fLastErrorMessage);
    end;
End;

function TSMSClient.getLastErrorMessage : string;
Begin
   Result := fLastErrorMessage;
End;

function TSMSClient.getLastErrorCode : integer;
Begin
   Result := fLastErrorCode;
End;

//***********************************************************************
//SMS Client events
//***********************************************************************


function smsFromPdu(pduhash : tozWideHash; fieldprefix : string) : TSMSMessage;
var
    sms : TSMSMessage;
Begin
      sms := TSMSMessage.Create;
      sms.messageID := pduhash.get(fieldprefix+'.messageid');
      sms.sender := pduhash.get(fieldprefix+'.originator');
      sms.receiver := pduhash.get(fieldprefix+'.recipient');
      sms.messagedata := pduhash.get(fieldprefix+'.messagedata');;
      sms.messagetype := pduhash.get(fieldprefix+'.messagetype');
      sms.serviceprovider := pduhash.get(fieldprefix+'.serviceprovider');
      sms.errorcode := strtoint(pduhash.get(fieldprefix+'.statuscode'));
      sms.errormessage := pduhash.get(fieldprefix+'.statusmessage');
      Result := sms;
End;

Type
myThread = Class(TThread)
public
smsS :TSMSMessage;
tS:TSMSClient;

Public
Procedure Execute; override;
End;

Procedure myThread.Execute;
Begin
tS.doOnMessageReceived(smsS);
Sleep(30);
End;


procedure TSMSClient.onIncomingPDU(inputPDU : string);
var
   pduhash : tozWideHash;
   action : string;
   errorcodestr : string;
   sms : TSMSMessage;
   transaction : string;
   deliverystatus : string;
   ackPDU : string;
   t: Tthread;
   o:Int32;
Begin
  pduhash := tozWideHash.Create;
  try
   pduhash.fromUrl(inputPDU);
   action := pduhash.get('action');
   if (action='error') then begin
      errorcodestr := pduhash.get('errorcode');
      fLastErrorCode := strtoint(errorcodestr);
      fLastErrorMessage := pduhash.get('errormessage');
      doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
   end else if (action='sendmessage') then begin
      //Sendmessage response
      sms := smsFromPdu(pduhash,'acceptreport');
      if (sms.errorcode<>0) then
         doOnMessageDeliveryError(sms)
      else
         doOnMessageAcceptedForDelivery(sms);
   end else if (action='deliveryreport') then begin
      //Acknowledge incoming delivery report
      transaction := pduhash.get('transactionid');
      ackPDU := stx+
      'transactionid='+transaction+'&'+
      'action=deliveryreport'+
      '&acceptreport.statuscode=0'+
      '&acceptreport.statusmessage=acknowledged'+
      etx;

      if (not sendPDU(ackPDU)) then begin
        doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
      end else begin
        //Incoming deliver report
        sms := smsFromPdu(pduhash,'deliveryreport');
        sms.sentdate  := pduhash.get('deliveryreport.deliveredtonetworkdate');
        sms.receiveddate  := pduhash.get('deliveryreport.deliveredtohandsetdate');
        deliverystatus := pduhash.get('deliveryreport.statuscode');
        if (deliverystatus='0') then

          // doOnMessageDeliveredToNetwork(sms)
        else if (deliverystatus='1') then
           doOnMessageDeliveredToHandset(sms)
        else
           doOnMessageDeliveryError(sms);

      end;
   end else if (action='messagereceived') then begin
      //Acknowledge incoming delivery report
      transaction := pduhash.get('transactionid');
      ackPDU := stx+
      'transactionid='+transaction+'&'+
      'action=messagereceived'+
      '&acceptreport.statuscode=0'+
      '&acceptreport.statusmessage=acknowledged'+
      etx;

      if (not sendPDU(ackPDU)) then begin
        doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
      end else begin

        //Incoming deliver report
        sms := smsFromPdu(pduhash,'message');
        sms.sentdate  := pduhash.get('message.sentdate');
        sms.receiveddate  := pduhash.get('message.receiveddate');

        if(bufferStop=BufferSize-2) then
        begin
            bufferStop:=0;
        end
        else
        begin
            bufferStop:=bufferStop+1;
        end;
             smsBuffer[bufferStop]:=sms;
      end;
   end else if (action='keepalive') then begin
      //do nothing
   end else begin
      fLastErrorCode := ERROR_PDU_UNEXPECTED;
      fLastErrorMessage := 'Unexpected data received from server. '+visualize(inputPDU);
      doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
   end;
   except

   end;
   pduhash.free;
End;



//***********************************************************************
//Processing thread management, windows messages
//***********************************************************************

procedure TSMSClient.startThread;
Begin
   processingThread := TProcessingThread.Create(true);
   processingThread.freeOnTerminate := false;
   processingThread.host := self;
   processingThread.synchronized := false;
   processingThread.resume;
End;

procedure TSMSClient.stopThread;
Begin
   processingThread.stop := true;
   postMessage(processingThread.msghandle,WM_OZ_STOP,0,0);
   processingThread.waitfor;
   processingThread.free;
   processingThread := nil;
End;

function TSMSClient.getMsgHandle : HWND;
Begin
   if (assigned(processingThread)) then Result := processingThread.msgHandle
   else Result := 0;
End;

function TDirectionToInt(direction : TDirection) : integer;
Begin
   if inbound=direction then Result := 0 else Result := 1;
End;

function IntToTDirection(i : integer) : TDirection;
Begin
   if i=0 then Result:= Inbound else Result := outbound;
End;


procedure TSMSClient.wndProc(var Msg: TMessage);
var
  inputPDU : string;
  cnt : integer;
  moredata : bool;
  p1 : pchar;
  commdata: string;
Begin
  case Msg.Msg of
     WM_OZ_STOP : begin
             //stopping thread
             end;
     WM_TCP_SOCKET_CONNECTED : begin
             //socket connected
             end;
     WM_CLIENT_CONNECTING : begin
             if (assigned(fOnConnecting)) then begin
                 try
                    fOnConnecting(self,fhost,fport);
                 except
                 end;
             end;
             end;
     WM_CLIENT_LINKUP : begin
             if (assigned(fOnSocketConnected)) then begin
                 try
                    fOnSocketConnected(self,host,port);
                 except
                 end;
             end;
             end;
     WM_CLIENT_CONNECTED: begin
             doOnClientConnected;
             end;
     WM_TCP_SOCKET_DISCONNECTED : begin
             if fConnected then doOnClientDisConnected;
             end;
     WM_TCP_DATAAVAILABLE : begin
             cnt := Msg.WParam;
             repeat
                moredata := readPDU(inputPDU,cnt);
                if (inputPDU<>'') then
                  onIncomingPDU(inputPDU);
             until not moredata;
             end;
     WM_CLIENT_MESSAGEDELIVERY_ERROR : begin
             doOnMessageDeliveryError(TSMSMessage(Msg.WParam));
             end;
     WM_CLIENT_KEEPALIVE : begin
             sendKeepalive;
             end;
     WM_DO_LOGIN : begin
             doLogin;
             end;
     WM_COMMUNICATION : begin
             p1 := pchar(Msg.LParam);
             if (assigned(fOnCommunication)) then begin
                commdata := p1;
                try
                fOnCommunication(Self,IntToTDirection(Msg.WParam),commdata);
                except
                end;
             end;
             freemem(p1);
             end;
     WM_CONNECT : begin
             beginAsyncConnect;
             end;
     end;

End;


//***********************************************************************
//Socket events - coming from a thread started by the socekt implementation
//***********************************************************************

procedure TSMSClient.onTCPClientConnected(Sender: TObject);
begin
   postMessage(processingThread.msghandle,WM_TCP_SOCKET_CONNECTED,0,0);
end;

procedure TSMSClient.onTCPClientDisConnected(Sender: TObject);
begin
   postMessage(processingThread.msghandle,WM_TCP_SOCKET_DISCONNECTED,0,0);
end;

procedure TSMSClient.onTCPClientStatus(axSender: TObject;
  const axStatus: TIdStatus; const asStatusText: String);
begin
  //Handle socket errors
end;

//***********************************************************************
//Socket control methods
//***********************************************************************
procedure TSMSClient.logCom(direction : TDirection; data : string);
var
   dsize : integer;
   p1 : pchar;
Begin
   dsize := length(data)+1;
   Getmem(p1,dsize);
   p1 := StrPCopy(p1, data);
   postMessage(processingThread.msghandle,WM_COMMUNICATION,TDirectionToInt(direction), integer(p1));
End;

function TSMSClient.sendPDU(pdu : string) : boolean;
var
   ret : bool;
begin
   try
       if (assigned(fOnCommunication)) then logCom(Outbound,copy(pdu,2,length(pdu)-2));

       tcpClient.Socket.WriteBufferOpen();
       tcpClient.Socket.Write(pdu);
       tcpClient.Socket.WriteBufferFlush();
       ret := true;
   except on e: exception do begin
       fLastErrorCode := ERROR_SOCKET_SEND;
       fLastErrorMessage := e.Message;
       disconnect;

       ret := false;
   end;
   end;
   Result := ret;
end;

function TSMSClient.waitforPDU(var pduReceived : string; timeout:integer):boolean;
var
   count,p : integer;
   tsnow,tsstart,tsexpire : int64;
   found : boolean;
   b:bool;
   ww:word;
   EncStream: Tmemorystream;
   bys:TBytes;
   sint:System.SmallInt;
   lenght:System.Integer;
Begin
   found := false;
   try
   tsstart := unixnow;
   tsexpire := tsstart+timeout;
   pduReceived := '';

   repeat

      tcpClient.Socket.ReadBytes(bys,-1,false);
     tcpInputBuffer:= TEncoding.ASCII.GetString(bys);

        p := pos(etx,tcpInputBuffer);
        if p>1 then begin  //min pdu length is 2 bytes: stx+etx
          pduReceived := copy(tcpInputBuffer,2,p-2);
          tcpInputBuffer := copy(tcpInputBuffer,p+1,length(tcpInputBuffer)-p);
          found := true;

          if (assigned(fOnCommunication))
          then logCom(Inbound,pduReceived);

        end;
     //end;
     if not found then sleep(100);
     tsnow  := unixnow;
   until ((found) or (tsexpire < tsnow));
   if (not found) then begin
       pduReceived := '';
       fLastErrorCode := ERROR_SOCKET_RECEIVE_TIMEOUT;
       fLastErrorMessage := 'No response received from server in '+inttostr(timeout)+' seconds.';
   end;
   except
       on e: exception do begin
           pduReceived := '';
           fLastErrorCode := ERROR_SOCKET_RECEIVE;
           fLastErrorMessage := e.Message;
       end;
   end;
   result := found;
End;

function TSMSClient.readPDU(var pduReceived : string; var count:integer) : boolean;
var
   p : integer;
   moremessages : boolean;
   tempstring : string;
   bys:TBytes;
Begin
   pduReceived := '';
   moremessages := false;
   try
     if count>0 then begin
      tcpClient.Socket.ReadBytes(bys,count,false);
      tcpInputBuffer:= tcpInputBuffer+TEncoding.ASCII.GetString(bys);
      count := 0;
      end;
      p := pos(etx,tcpInputBuffer);
      if p>1 then begin  //min pdu length is 2 bytes: stx+etx
          tempstring := copy(tcpInputBuffer,2,p-2);
          pduReceived := tempstring;

          if (assigned(fOnCommunication)) then logCom(Inbound,pduReceived);
          tcpInputBuffer := copy(tcpInputBuffer,p+1,length(tcpInputBuffer)-p);

          p := pos(etx,tcpInputBuffer);
          if (p>1) then moremessages := true;
      end;
   except
       on e: exception do begin
           pduReceived := '';
           fLastErrorCode := ERROR_SOCKET_RECEIVE;
           fLastErrorMessage := e.Message;
           doonClientConnectionError(fLastErrorCode,fLastErrorMessage);
       end;
   end;
   Result:=moremessages;
End;

//***********************************************************************
//Event handler wrappers
//***********************************************************************

procedure TSMSClient.doOnClientConnected;
Begin
    if (assigned(OnClientConnected)) then begin
        try
            fOnClientConnected(self);
        except
        end;
    end;
End;

procedure TSMSClient.doOnClientDisconnected;
Begin
    fConnected := false;
    if (assigned(OnClientDisconnected)) then begin
        try
            fOnClientDisconnected(self);
        except
        end;
    end;
End;

procedure TSMSClient.doOnMessageAcceptedForDelivery(sms : TSMSMessage);
Begin
    if (assigned(OnMessageAcceptedForDelivery)) then begin
        try
           fOnMessageAcceptedForDelivery(self,sms.messageid,sms.sender,sms.receiver,sms.messagedata,sms.messagetype,
            sms.serviceprovider,sms.sentdate,sms.receiveddate);
        except
        end;
    end;
    sms.Free;
End;

procedure TSMSClient.doOnMessageDeliveredToNetwork(sms : TSMSMessage);
Begin
    if (assigned(OnMessageDeliveredToNetwork)) then begin
        try
           fOnMessageDeliveredToNetwork(self,sms.messageid,sms.sender,sms.receiver,sms.messagedata,sms.messagetype,
            sms.serviceprovider,sms.sentdate,sms.receiveddate)
        except
        end;
    end;
    sms.Free;
End;



procedure TSMSClient.doOnMessageDeliveredToHandset(sms : TSMSMessage);
Begin
    if (assigned(OnMessageDeliveredToHandset)) then begin
        try
            fOnMessageDeliveredToHandset(self,sms.messageid,sms.sender,sms.receiver,sms.messagedata,sms.messagetype,
            sms.serviceprovider,sms.sentdate,sms.receiveddate);
        except
        end;
    end;
    sms.Free;
End;

procedure TSMSClient.doOnMessageDeliveryError(sms : TSMSMessage);
Begin
    if (assigned(OnMessageDeliveryError)) then begin
        try
            fOnMessageDeliveryError(self,sms.messageid,sms.sender,sms.receiver,sms.messagedata,sms.messagetype,
            sms.serviceprovider,sms.sentdate,sms.receiveddate,sms.errorCode,sms.errorMessage);
        except
        end;
    end;
    sms.Free;
End;

procedure TSMSClient.doOnMessageReceived(sms : TSMSMessage);
Begin
    if (assigned(OnMessageReceived)) then begin
        try
            fOnMessageReceived(self,sms.messageid,sms.sender,sms.receiver,sms.messagedata,sms.messagetype,
            sms.serviceprovider,sms.sentdate,sms.receiveddate);
        except
        end;
    end;
    sms.Free;
End;

procedure TSMSClient.doonClientConnectionError(errorCode: integer; errorMessage:string);
Begin
    fConnected := false;
    if (assigned(onClientConnectionError)) then begin
        disconnect;
        try
            fonClientConnectionError(self,errorCode,errorMessage);
        except
        end;
    end;

end;

//***********************************************************************
//Support functions
//***********************************************************************
function TSMSClient.ozUrlEncode(data : widestring) : string;
var utf8str : String;
Begin
   utf8str := utf8encode(data);
   Result := TIdURI.ParamsEncode(utf8str);
End;

function TSMSClient.DateTimeToUnix(ConvDate: TDateTime): Longint;
const
  UnixStartDate: TDateTime = 25569.0;
begin
  Result := Round((ConvDate - UnixStartDate) * 86400);
end;

function TSMSClient.unixnow : int64;
Begin
 Result := DateTimeToUnix(now);
End;

function TSMSClient.randomStringL(len : integer):string;
var
  s : string;
  x : integer;
Begin
  for x := 1 to len do begin
     s := s+chr(ord('A')+random(26));
  end;
  Result := s;
End;

function TSMSClient.visualize(s : string):string;
var ret : string;
Begin
    ret := AnsiReplaceStr(s,stx,'<stx>');
    ret := AnsiReplaceStr(ret,etx,'<etx>');
    Result := ret;
End;

function TSMSClient.getNextTransID : string;
Begin
   transactionID := transactionID+2;
   if (transactionID>32768) then transactionID := 1;
   Result := inttostr(transactionID);
End;

procedure TSMSClient.onKeepalive(Sender: TObject);
Begin
   postMessage(processingThread.msghandle,WM_CLIENT_KEEPALIVE,0,0);
End;

procedure Register;
begin
  RegisterComponents('Ozeki', [TSMSClient]);
end;

end.
